//
// pvjpeg.c - Routines for JPEG support in Purpleview.
// Uses the untitled software of the Independent JPEG Group.
//

#include "pvconf.h"

#if PV_JPEG_SUPPORT

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <jpeglib.h>
#include <jerror.h>

#include "purpleview.h"
#include <strsafe.h>

#if BITS_IN_JSAMPLE != 8
#error JSAMPLE bits must be 8
#endif


struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */
  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

static void
my_error_exit (j_common_ptr cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}


int pv_jpeg_open(HWND hwnd,
					  struct iinfo_struct *iinfo,
					  struct einfo_struct *einfo)
{

	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;

	FILE *f;
	int retval;
	unsigned char *row;
	LPBITMAPINFOHEADER lpdib;  // could use info->lpdib, but syntax gets messy
	unsigned char *lpdib_b; // alias of lpdib
	unsigned char *lpbits;
	RGBQUAD* palette;
	int grayscale;
	int dib_size;
	int i,j;
	int bytesperrow;
	JSAMPROW row_pointer[1];	/* pointer to JSAMPLE row[s] */
	int jpeg_bytesperrow;
	int cinfo_valid;
	struct _stat stbuf;

	retval=0;
	cinfo_valid=0;
	row = NULL;
	lpdib = NULL;


	i=sizeof(struct jpeg_decompress_struct);

	// fixme   why is this here??
	if(_tstat( iinfo->img_filename, &stbuf ) ==0) {
		iinfo->file_size = stbuf.st_size;
	}

	f=_tfopen(iinfo->img_filename,_T("rb"));
	if(!f) {
		pv_mesg(hwnd,MSG_E,_T("Can't open %s"),iinfo->img_filename);
		return 0;
	}

	cinfo.err = jpeg_std_error(&jerr.pub);

	jerr.pub.error_exit = my_error_exit;
	if (setjmp(jerr.setjmp_buffer)) {
		char buffer[JMSG_LENGTH_MAX];

		/* Create the message */
		(*cinfo.err->format_message) ((j_common_ptr)&cinfo, buffer);

#ifdef UNICODE
		pv_mesg(hwnd,MSG_E,_T("libjpeg reports error:\r\n%S"),buffer);
#else
		pv_mesg(hwnd,MSG_E,"libjpeg reports error:\r\n%s",buffer);
#endif

		goto done;
	}

	jpeg_create_decompress(&cinfo);
	cinfo_valid=1;

	jpeg_stdio_src(&cinfo, f);

	jpeg_read_header(&cinfo, TRUE);

	iinfo->res_flag=0; // default to unknown
	iinfo->res_x =  (int) cinfo.X_density;
	iinfo->res_y =  (int) cinfo.Y_density;

	if(iinfo->res_x>0 && iinfo->res_y>0) {
		if(cinfo.density_unit==1) { // dots/inch
			iinfo->res_x = (int) (((double)iinfo->res_x)*39.37);
			iinfo->res_y = (int) (((double)iinfo->res_y)*39.37);
			iinfo->res_flag=1; // converted to meters
		}
		else if(cinfo.density_unit==2) { // dots/cm
			iinfo->res_x *= 100;
			iinfo->res_y *= 100;
			iinfo->res_flag=1; // converted to meters
		}
		else {
			iinfo->res_flag=2;  // unknown units
		}
	}

	switch(cinfo.jpeg_color_space) {
	case JCS_GRAYSCALE: iinfo->colortype= COLORTYPE_GRAY; break;
	case JCS_RGB:   iinfo->colortype= COLORTYPE_RGB; break;
	case JCS_YCbCr: iinfo->colortype= COLORTYPE_YCbCr; break;
	case JCS_CMYK:  iinfo->colortype= COLORTYPE_CMYK; break;
	case JCS_YCCK:  iinfo->colortype= COLORTYPE_YCCK; break;
	default: iinfo->colortype= COLORTYPE_UNKNOWN;
	}


	// I don't know if this is the right way to detect
	// a progressive JPEG. It doesn't seem to be documented.
	iinfo->interlaced = (cinfo.progressive_mode)?1:0;

	jpeg_start_decompress(&cinfo);

	if(cinfo.output_components!=1 && cinfo.output_components!=3
		&& !(cinfo.output_components==4 && iinfo->colortype==COLORTYPE_YCCK))
	{
		pv_mesg(hwnd,MSG_E,_T("This type of JPEG is not supported"));
		goto done;
	}
	grayscale = (cinfo.output_components==1);

	if(grayscale) {
		bytesperrow = pv_get_dib_bytes_per_row(8, cinfo.output_width);
		dib_size = sizeof(BITMAPINFOHEADER) + 256*sizeof(RGBQUAD) +
			cinfo.output_height*bytesperrow;
	}
	else {
		bytesperrow = pv_get_dib_bytes_per_row(24, cinfo.output_width);
		dib_size = sizeof(BITMAPINFOHEADER) + cinfo.output_height*bytesperrow;
	}

	lpdib = (LPBITMAPINFOHEADER)HeapAlloc(einfo->heap,HEAP_ZERO_MEMORY,dib_size);
	if(!lpdib) {
		pv_mesg(hwnd,MSG_E,_T("Can't malloc"));
		goto done;
	}
	lpdib_b=(unsigned char*)lpdib;

	lpdib->biSize = sizeof(BITMAPINFOHEADER);

	iinfo->bitsperpixel = 8*cinfo.output_components;

	if(grayscale) {
		palette = (RGBQUAD*) &lpdib_b[lpdib->biSize];
		for(i=0;i<256;i++) {
			palette[i].rgbRed=i;
			palette[i].rgbGreen=i;
			palette[i].rgbBlue=i;
			palette[i].rgbReserved=0;
		}
		lpbits= &lpdib_b[lpdib->biSize+1024];
		lpdib->biClrUsed = 256;
	}
	else {
		lpbits= &lpdib_b[lpdib->biSize];
		lpdib->biClrUsed = 0;
	}

	iinfo->palettecolors = 0;
	iinfo->compression = COMPR_JPEG;
	iinfo->img_fmt = FMT_JPEG;
	iinfo->has_bgcolor = 0;

	lpdib->biWidth = cinfo.output_width;
	lpdib->biHeight = cinfo.output_height;
	lpdib->biPlanes = 1;
	lpdib->biBitCount = grayscale?8:24; //iinfo->bitsperpixel;
	lpdib->biCompression = BI_RGB;
	lpdib->biSizeImage = cinfo.output_height*bytesperrow;

	if(iinfo->res_flag==1) {
		lpdib->biXPelsPerMeter=iinfo->res_x;
		lpdib->biYPelsPerMeter=iinfo->res_y;
	}
	else {
		lpdib->biXPelsPerMeter = 0;
		lpdib->biYPelsPerMeter = 0;
	}

	lpdib->biClrImportant = 0;

	jpeg_bytesperrow= cinfo.output_width * cinfo.output_components;
	row = malloc(jpeg_bytesperrow);
	if(!row) {
		pv_mesg(hwnd,MSG_E,_T("Can't malloc"));
		goto done;
	}

	row_pointer[0] = row;
	while (cinfo.output_scanline < cinfo.output_height) {
		j=cinfo.output_scanline;
		jpeg_read_scanlines(&cinfo, row_pointer, 1);

		// copy to DIB
		if(grayscale) {
			for(i=0;i<(int)cinfo.output_width;i++) {
				lpbits[bytesperrow*(cinfo.output_height-1-j)+i]=row[i];
			}
		}
		else if(cinfo.output_components==4) { // YCCK
			for(i=0;i<(int)cinfo.output_width;i++) {
				unsigned char *tmpc;
				unsigned char *kk;
				float r,g,b;

				tmpc=&lpbits[bytesperrow*(cinfo.output_height-1-j)+i*3];
				kk = &row[i*4];

				r=  32.2322F  +0.847486F *kk[0]                  +0.0212052F*kk[2] -0.44168F *(255-kk[3]);
				g=  -6.45922F +0.302768F *kk[0] +0.632835F*kk[1] +0.0800831F*kk[2] -0.412053F*(255-kk[3]);
				b= -30.2647F  +0.0906424F*kk[0] +0.383431F*kk[1] +0.64052F  *kk[2] -0.397349F*(255-kk[3]);

				if(r<0.0F)         tmpc[2] = 0;
				else if(r>255.0F)  tmpc[2] = 255;
				else               tmpc[2] = (unsigned char) r;

				if(g<0.0F)         tmpc[1] = 0;
				else if(g>255.0F)  tmpc[1] = 255;
				else               tmpc[1] = (unsigned char)g;

				if(b<0.0F)         tmpc[0] = 0;
				else if(b>255.0F)  tmpc[0] = 255;
				else               tmpc[0] = (unsigned char)b;

			}
		}
		else {
			for(i=0;i<(int)cinfo.output_width;i++) {
				unsigned char *tmpc;
				tmpc=&lpbits[bytesperrow*(cinfo.output_height-1-j)+i*3];
				tmpc[2]=row[i*3+0]; // R
				tmpc[1]=row[i*3+1]; // G
				tmpc[0]=row[i*3+2]; // B
			}
		}
	}

	jpeg_finish_decompress(&cinfo);
	retval=1;

done:
	if(cinfo_valid) jpeg_destroy_decompress(&cinfo);
	if(f) fclose(f);

	if(lpdib && !retval) HeapFree(einfo->heap,0,(LPVOID)lpdib);
	if(row) free(row);

	if(retval) {
		iinfo->lpdib = lpdib;
		iinfo->lpdib_size = dib_size;
	}

	return retval;
}


int pv_jpeg_get_copyright(TCHAR *buf, int buflen)
{
	struct jpeg_error_mgr jerr;
	const char *jversion;
	const char *jcopyright;
	TCHAR buf2[500];

	StringCchCopy(buf,buflen,_T("This software is based in part on the work ")
		_T("of the Independent JPEG Group.\r\n"));

	jpeg_std_error(&jerr);
	jversion = jerr.jpeg_message_table[JMSG_VERSION];
	jcopyright = jerr.jpeg_message_table[JMSG_COPYRIGHT];
#ifdef UNICODE
	StringCbPrintf(buf2,sizeof(buf2),_T("libjpeg version %S\r\n%S\r\n\r\n"),jversion,jcopyright);
#else
	StringCbPrintf(buf2,sizeof(buf2),"libjpeg version %s\r\n%s\r\n\r\n",jversion,jcopyright);
#endif
	StringCchCat(buf,buflen,buf2);
	return 1;
}


#endif  // PV_JPEG_SUPPORT
